@@ -157,6 +157,10 @@ public class MimeMailer extends Mailer {
                 // SMTP provider
                 props.put("mail.smtp.socketFactory.class", SSL_FACTORY);
                 props.put("mail.smtp.socketFactory.fallback", "false");
+                if (isPortExplicitlySpecified()) {
+                    props.put("mail.smtp.socketFactory.port",
+                              String.valueOf(port));
+                }
             }
             if (user != null || password != null) {
                 props.put("mail.smtp.auth", "true");
@@ -285,9 +289,10 @@ public class MimeMailer extends Mailer {
 
     private static InternetAddress[] internetAddresses(Vector list)
         throws AddressException, UnsupportedEncodingException {
-        InternetAddress[] addrs = new InternetAddress[list.size()];
+        final int size = list.size();
+        InternetAddress[] addrs = new InternetAddress[size];
 
-        for (int i = 0; i < list.size(); ++i) {
+        for (int i = 0; i < size; ++i) {
             EmailAddress addr = (EmailAddress) list.elementAt(i);
 
             String name = addr.getName();
