@@ -54,8 +54,10 @@ import org.apache.tools.ant.util.VectorSet;
 
 public class FTPTaskMirrorImpl implements FTPTaskMirror {
 
-    /** return code of ftp - not implemented in commons-net version 1.0 */
+    /** return code of ftp */
     private static final int CODE_521 = 521;
+    private static final int CODE_550 = 550;
+    private static final int CODE_553 = 553;
 
     /** Date formatter used in logging, note not thread safe! */
     private static final SimpleDateFormat TIMESTAMP_LOGGING_SDF =
@@ -722,7 +724,8 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
                     throw new BuildException("could not change working dir to "
                                              + parent.curpwd);
                 }
-                for (int fcount = 0; fcount < pathElements.size() - 1; fcount++) {
+                final int size = pathElements.size();
+                for (int fcount = 0; fcount < size - 1; fcount++) {
                     String currentPathElement = (String) pathElements.elementAt(fcount);
                     try {
                         boolean result = this.client.changeWorkingDirectory(currentPathElement);
@@ -736,7 +739,7 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
                         } else if (!result) {
                             return;
                         }
-                        this.curpwd = this.curpwd + task.getSeparator()
+                        this.curpwd = getCurpwdPlusFileSep()
                             + currentPathElement;
                     } catch (IOException ioe) {
                         throw new BuildException("could not change working dir to "
@@ -745,7 +748,7 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
                     }
 
                 }
-                String lastpathelement = (String) pathElements.elementAt(pathElements.size() - 1);
+                String lastpathelement = (String) pathElements.elementAt(size - 1);
                 FTPFile [] theFiles = listFiles(this.curpwd);
                 this.ftpFile = getFile(theFiles, lastpathelement);
             }
@@ -797,7 +800,7 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
              * @return absolute path as string
              */
             public String getAbsolutePath() {
-                return curpwd + task.getSeparator() + ftpFile.getName();
+                return getCurpwdPlusFileSep() + ftpFile.getName();
             }
             /**
              * find out the relative path assuming that the path used to construct
@@ -850,7 +853,8 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
                 Vector pathElements2 = SelectorUtils.tokenizePath(currentPath,
                                                                   task.getSeparator());
                 String relPath = currentRelativePath;
-                for (int pcount = pathElements2.size(); pcount < pathElements.size(); pcount++) {
+                final int size = pathElements.size();
+                for (int pcount = pathElements2.size(); pcount < size; pcount++) {
                     String currentElement = (String) pathElements.elementAt(pcount);
                     FTPFile[] theFiles = listFiles(currentPath);
                     FTPFile theFile = null;
@@ -942,6 +946,17 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
             public String getCurpwd() {
                 return curpwd;
             }
+            /**
+             * returns the path of the directory containing the AntFTPFile.
+             * of the full path of the file itself in case of AntFTPRootFile
+             * and appends the remote file separator if necessary.
+             * @return parent directory of the AntFTPFile
+             * @since Ant 1.8.2
+             */
+            public String getCurpwdPlusFileSep() {
+                String sep = task.getSeparator();
+                return curpwd.endsWith(sep) ? curpwd : curpwd + sep;
+            }
             /**
              * find out if a symbolic link is encountered in the relative path of this file
              * from rootPath.
@@ -1216,7 +1231,8 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
             throw new BuildException("at least one fileset must be specified.");
         } else {
             // get files from filesets
-            for (int i = 0; i < task.getFilesets().size(); i++) {
+            final int size = task.getFilesets().size();
+            for (int i = 0; i < size; i++) {
                 FileSet fs = (FileSet) task.getFilesets().elementAt(i);
 
                 if (fs != null) {
@@ -1761,8 +1777,9 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
                     //  to indicate that an attempt to create a directory has
                     //  failed because the directory already exists.
                     int rc = ftp.getReplyCode();
-                    if (!(task.isIgnoreNoncriticalErrors() && (rc == FTPReply.CODE_550
-                                                               || rc == FTPReply.CODE_553 || rc == CODE_521))) {
+                    if (!(task.isIgnoreNoncriticalErrors() && (rc == CODE_550
+                                                               || rc == CODE_553
+                                                               || rc == CODE_521))) {
                         throw new BuildException("could not create directory: "
                                                  + ftp.getReplyString());
                     }
@@ -1791,8 +1808,9 @@ public class FTPTaskMirrorImpl implements FTPTaskMirror {
     private void handleMkDirFailure(FTPClient ftp)
         throws BuildException {
         int rc = ftp.getReplyCode();
-        if (!(task.isIgnoreNoncriticalErrors() && (rc == FTPReply.CODE_550
-                                                   || rc == FTPReply.CODE_553 || rc == CODE_521))) {
+        if (!(task.isIgnoreNoncriticalErrors() && (rc == CODE_550
+                                                   || rc == CODE_553
+                                                   || rc == CODE_521))) {
             throw new BuildException("could not create directory: "
                                      + ftp.getReplyString());
         }
