@@ -27,6 +27,7 @@ import java.util.Collections;
 import java.util.Date;
 import java.util.Enumeration;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Map;
 import java.util.zip.CRC32;
 import java.util.zip.Inflater;
@@ -182,8 +183,8 @@ public class ZipFile {
         archive = new RandomAccessFile(f, "r");
         boolean success = false;
         try {
-            Map entriesWithoutEFS = populateFromCentralDirectory();
-            resolveLocalFileHeaderData(entriesWithoutEFS);
+            Map entriesWithoutUTF8Flag = populateFromCentralDirectory();
+            resolveLocalFileHeaderData(entriesWithoutUTF8Flag);
             success = true;
         } finally {
             if (!success) {
@@ -269,7 +270,13 @@ public class ZipFile {
                 return bis;
             case ZipEntry.DEFLATED:
                 bis.addDummy();
-                return new InflaterInputStream(bis, new Inflater(true));
+                final Inflater inflater = new Inflater(true);
+                return new InflaterInputStream(bis, inflater) {
+                    public void close() throws IOException {
+                        super.close();
+                        inflater.end();
+                    }
+                };
             default:
                 throw new ZipException("Found unsupported compression method "
                                        + ze.getMethod());
@@ -308,7 +315,7 @@ public class ZipFile {
      */
     private Map populateFromCentralDirectory()
         throws IOException {
-        HashMap noEFS = new HashMap();
+        HashMap noUTF8Flag = new HashMap();
 
         positionAtCentralDirectory();
 
@@ -334,10 +341,10 @@ public class ZipFile {
             off += SHORT; // skip version info
 
             final int generalPurposeFlag = ZipShort.getValue(cfh, off);
-            final boolean hasEFS = 
-                (generalPurposeFlag & ZipOutputStream.EFS_FLAG) != 0;
+            final boolean hasUTF8Flag = 
+                (generalPurposeFlag & ZipOutputStream.UFT8_NAMES_FLAG) != 0;
             final ZipEncoding entryEncoding =
-                hasEFS ? ZipEncodingHelper.UTF8_ZIP_ENCODING : zipEncoding;
+                hasUTF8Flag ? ZipEncodingHelper.UTF8_ZIP_ENCODING : zipEncoding;
 
             off += SHORT;
 
@@ -400,11 +407,11 @@ public class ZipFile {
             archive.readFully(signatureBytes);
             sig = ZipLong.getValue(signatureBytes);
 
-            if (!hasEFS && useUnicodeExtraFields) {
-                noEFS.put(ze, new NameAndComment(fileName, comment));
+            if (!hasUTF8Flag && useUnicodeExtraFields) {
+                noUTF8Flag.put(ze, new NameAndComment(fileName, comment));
             }
         }
-        return noEFS;
+        return noUTF8Flag;
     }
 
     private static final int MIN_EOCD_SIZE =
@@ -445,12 +452,16 @@ public class ZipFile {
         throws IOException {
         boolean found = false;
         long off = archive.length() - MIN_EOCD_SIZE;
-        long stopSearching = Math.max(0L, archive.length() - MAX_EOCD_SIZE);
+        final long stopSearching =
+            Math.max(0L, archive.length() - MAX_EOCD_SIZE);
         if (off >= 0) {
-            archive.seek(off);
-            byte[] sig = ZipOutputStream.EOCD_SIG;
-            int curr = archive.read();
-            while (off >= stopSearching && curr != -1) {
+            final byte[] sig = ZipOutputStream.EOCD_SIG;
+            for (; off >= stopSearching; off--) {
+                archive.seek(off);
+                int curr = archive.read();
+                if (curr == -1) {
+                    break;
+                }
                 if (curr == sig[POS_0]) {
                     curr = archive.read();
                     if (curr == sig[POS_1]) {
@@ -464,8 +475,6 @@ public class ZipFile {
                         }
                     }
                 }
-                archive.seek(--off);
-                curr = archive.read();
             }
         }
         if (!found) {
@@ -499,9 +508,9 @@ public class ZipFile {
      * <p>Also records the offsets for the data to read from the
      * entries.</p>
      */
-    private void resolveLocalFileHeaderData(Map entriesWithoutEFS)
+    private void resolveLocalFileHeaderData(Map entriesWithoutUTF8Flag)
         throws IOException {
-        Enumeration e = getEntries();
+        Enumeration e = Collections.enumeration(new HashSet(entries.keySet()));
         while (e.hasMoreElements()) {
             ZipEntry ze = (ZipEntry) e.nextElement();
             OffsetEntry offsetEntry = (OffsetEntry) entries.get(ze);
@@ -531,10 +540,15 @@ public class ZipFile {
             offsetEntry.dataOffset = offset + LFH_OFFSET_FOR_FILENAME_LENGTH
                 + SHORT + SHORT + fileNameLen + extraFieldLen;
 
-            if (entriesWithoutEFS.containsKey(ze)) {
+            if (entriesWithoutUTF8Flag.containsKey(ze)) {
+                // changing the name of a ZipEntry is going to change
+                // the hashcode
+                // - see https://issues.apache.org/jira/browse/COMPRESS-164
+                entries.remove(ze);
                 setNameAndCommentFromExtraFields(ze,
                                                  (NameAndComment)
-                                                 entriesWithoutEFS.get(ze));
+                                                 entriesWithoutUTF8Flag.get(ze));
+                entries.put(ze, offsetEntry);
             }
         }
     }
